/**
 *#########################################################################
 * BrowseResponseData.java - part of the demo-client for Greenstone 3, 
 * of the Greenstone digital library suite from the New Zealand Digital 
 * Library Project at the  * University of Waikato, New Zealand.
 * <BR><BR>
 * Copyright (C) 2008 New Zealand Digital Library Project
 * <BR><BR>
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * <BR><BR>
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *########################################################################
 */

package org.greenstone.gs3client.data;

import java.util.Map;
import org.greenstone.gsdl3.util.GSXML;
import org.w3c.dom.Element;

/** 
 * Stores the data contained in an XML response message to a request for 
 * the classification Browse structure. An object of this class can be reused
 * after instatiation by calling setResponseData() with a new browse
 * response XML message. This will first call clear() to clear/release 
 * its references to all the old data.
 * @author ak19 
*/
public class BrowseResponseData extends ResponseData {
	// Inherits Map nodeIDsToNodes of (nodeID, NodeData ref) pairs which
	// maintains the NodeData object refs in order of their insertion into 
	// the Map (LinkedHashMap).
	// This class keeps track of both ClassifierNode and DocumentNode data
	// objects that are returned in response to a browse message request.
	
	/** Reference to the root browse classifier, which may contain further
	 * descendent classifierNodes or documentNodes. */
	protected ClassifierNodeData rootClassifierNode;
	
	/** Default constructor */
	public BrowseResponseData() {
		super();
		rootClassifierNode = null;
	}
	
	/** Resets the internal data members of this BrowseResponseData object of 
	 * their values so that this BrowseResponseData can be reused for the
	 * next Browse response message. */
	public void clear() {
		super.clear(); 
			// clears Map nodeIDsToNodes of (nodeID, NodeData ref) pairs 
		rootClassifierNode = null;
		System.gc();
	}
	
	/** Given an XML responseMessage to a request for the classification  
	 * Browse structure, this method sets its nodeIDsToNodes member variable 
	 * by instantiating classifierNodeData and/or documentNodeData objects   
	 * as given in the responseMessageTag. This method will not process all 
	 * descendents of a classifierNode, but only those that are children
	 * of the top level classifierNode. 
	 * The responseMessageTag contains *all the descendents* in one go. But
	 * we don't set all the descendents, just the children of the 
	 * rootClassifierNode. However, as we store the responseMsgTag in pieces
	 * (each child NodeData stores the bit relevant to it), we will work out
	 * the children of each child etc as needed.
	 * @param responseMsgTag is the the XML DOM Element representing a browse 
	 * response XML message. */
	public void setResponseData(Element responseMsgTag) 
	{	
		this.clear();
		// Now process the XML element responseMsgTag:
		// First go to the nodeStructureTag in the response: should only
		// be one, then get the first <classifierNode> therein - 
		// (its nodeID should match with this.classifier.name, i.e.
		// the element should refer to the rootClassifierNode.
		Element nodeStructTag 
			= ParseUtil.getFirstDescElementCalled(
					responseMsgTag, GSXML.NODE_STRUCTURE_ELEM);
		Element rootClassifierTag = ParseUtil.getFirstChildElementCalled(
				nodeStructTag, GSXML.CLASS_NODE_ELEM);
	
		// create a ClassifierNodeData for the classifier object which serves
		// as the Root of the entire classification structure:
		this.rootClassifierNode = new ClassifierNodeData(rootClassifierTag);
		
		
		if(rootClassifierNode != null) {
			// Don't forget to put the root into the map as well:
			this.nodeIDsToNodes.put(rootClassifierNode.nodeID, 
					rootClassifierNode); 
			
			// Now obtain all *child elements* of this element (actually
			// these will be children of the rootClassifierNode) that are   
			// either <documentNode>s or <classifierNodes>	
			this.rootClassifierNode.setChildren(this.nodeIDsToNodes);
		}
	} 
	
	/** Given the response message element to a metadata request on
	 * &lt;classifierNode&gts;, this method sets those classifierNodes' metadata. 
	 * It merely calls the the superclass' setMetadataForDocuments(messageTag, 
	 * tagName) method to parse out data for the tagName=&lt;classifierNode&gt;. 
	 * @param messageTag - the response message Element to a metadata
	 * request on &lt;classifierNode&gt;s */
	public boolean setMetadataForClassifiers(Element messageTag) {
		return setMetadataForNodes(messageTag, GSXML.CLASS_NODE_ELEM);
	}
	
	/** @return the ClassifierNode representing the top-level/root 
	 * &lt;classifierNode&gt; in the Browse response. */
	public ClassifierNodeData getRootClassifier() { 
		return rootClassifierNode; 
	}
	
	/** @return the hashMap of mappings from nodeIDs to DocumentNodeData 
	 * object references */
	public Map getIDToNodeMapping() { return nodeIDsToNodes; }
	
	/** Overloaded method to produce a String output of the data contained in 
	 * this BrowseResponseMessage. It merely calls rootClassifierNode.show()
	 * @return a String representation of the data in this BrowseResponseMessage */
	public String show() {
		return this.rootClassifierNode.show();
	}
}