/**
 *#########################################################################
 * CollectionData.java - part of the demo-client for Greenstone 3, of the 
 * Greenstone digital library suite from the New Zealand Digital Library 
 * Project at the  * University of Waikato, New Zealand.
 * <BR><BR>
 * Copyright (C) 2008 New Zealand Digital Library Project
 * <BR><BR>
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * <BR><BR>
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *########################################################################
 */

package org.greenstone.gs3client.data;

import org.greenstone.gsdl3.util.GSXML;
import org.w3c.dom.Element;
import org.w3c.dom.NodeList;

import java.util.Vector;
import java.util.HashMap;


/**
 * CollectionData represents a Greenstone 3 collection and contains the data
 * stored in a &lt;collection&gt;&lt;/collection&gt; element returned from a   
 * MessageRouter's desribe response XML and a Collection describe response XML. 
 * A CollectionData object can therefore contain services (represented by 
 * ServiceData), a list of collection-level metadata (represented by MetaData) 
 * and a list of plugins (represented by PluginData). 
 * 
 * ServiceData and PluginData are static inner classes of CollectionData. 
 * To declare variables of these inner class types in other java files/classes,
 * import each one: e.g. "import gs3client.CollectionData.ServiceData".
 * Then they can be used without the outerclass' name for qualification:
 * i.e. can then use Servicedata sv = new ServiceData(...)
 * instead of CollectionData.ServiceData sv = new
 * CollectionData.ServiceData(...).
 * @author ak19
*/
public class CollectionData {
	/** Stores the location of images and stuff. Not really used at present */
	protected String url = ""; 
	
	/** The basic field of a collection: the name of the collection. This is
	 * present for collection elements in all describe response messages,
	 * whether sent from the MessageRouter or from Greenstone collections. */
	public final String name;
	
	/* display items - these fields are set later with setFields() */
	/** Display name for the Greenstone collection represented by this 
	 * CollectionData */
	protected String displayName;
	/** Description of the Greenstone collection represented by this 
	 * CollectionData */
	protected String description;
	
	protected String smallIcon;
	protected String icon;
	
	/** List of services supported by this collection */
	protected ServiceData[] serviceList = null;
	/** Collection level metadata */
	protected MetaData[] metadataList = null;
	/** Greenstone plugins used to build this collection with */
	protected PluginData[] pluginList = null;
	
	/** 
	 * Gets a &lt;collection&gt;&lt;/collection&gt; element as returned by the 
	 * <b>MessageRouter's</b> describe response, with only the name field set.
	 * @param mrCollectionTag the &lt;collection&gt;&lt;/collection&gt; element  
	 * as returned by a describe response from the <b>MessageRouter</b>.
	 */
	public CollectionData(Element mrCollectionTag){
		// <collection name="NameValue"></collection> - retrieve NameValue first
		// We'd definitely have a name attribute, but to be on the safe side:
		this.name = mrCollectionTag.hasAttribute(GSXML.NAME_ATT) ?
			mrCollectionTag.getAttribute(GSXML.NAME_ATT) : "";
	}
	
	/** 
	 * Gets a &lt;collection&gt;&lt;/collection&gt; element - as returned by a 
	 * describe response from a <b>Collection</b> - and sets the member vars using 
	 * the data in there.
	 * @param collectionTag is the &lt;collection&gt;&lt;/collection&gt; element  
	 * as returned by a describe response from a <b>Collection</b>.
	 */
	public void setFields(Element collectionTag) {
		// Get the basic attributes: located in <displayItem> children
		// of collectionTag:
		HashMap map = ParseUtil.getElementValuesForAttr(collectionTag,
				GSXML.DISPLAY_TEXT_ELEM, GSXML.DISPLAY_TEXT_NAME);
		// Variable map now contains all pairs 
		// (name-attribute's value, displayItem's bodytext) 
		// extracted from <displayItem>s of the parent collectionTag element
		String temp = (String)map.get("name"); 
		this.displayName = (temp == null || temp.equals("")) ? 
				this.name : temp;
		temp = (String)map.get("description");
		this.description = (temp == null) ? "" : temp;
		temp = (String)map.get("smallicon");
		this.smallIcon = (temp == null) ? "" : temp;
		temp = (String)map.get("icon");
		this.icon = (temp == null) ? "" : temp;
		map.clear();
		map = null;
		
		// Now get the serviceList, metadataList, pluginList
		// (1) Get all the <service></service> in <serviceList> 
		Vector v = ParseUtil.getListElementsAsArray(collectionTag, 
				GSXML.SERVICE_ELEM+GSXML.LIST_MODIFIER, GSXML.SERVICE_ELEM);
		if(v != null) { // v contains individual <service></service> elements
			serviceList = new ServiceData[v.size()];
			for(int i = 0; i < serviceList.length; i++)
				serviceList[i] = new ServiceData((Element)v.get(i));
			v.clear();
		}
		
		// (2) Get all the <metadata></metadata> elements in <metadataList>
		v = ParseUtil.getListElementsAsArray(collectionTag, 
				GSXML.METADATA_ELEM+GSXML.LIST_MODIFIER, GSXML.METADATA_ELEM);
		if(v != null) { // v contains individual <service></service> elements
			metadataList = new MetaData[v.size()];
			for(int i = 0; i < metadataList.length; i++)
				metadataList[i] = new MetaData((Element)v.get(i));
			v.clear();
		}
		
		// (3) Get all the <plugin></plugin> elements in <metadataList>
		// Is it necessary for the Client to keep track of the plugins 
		// in any collection?
		
		v = ParseUtil.getListElementsAsArray(collectionTag, 
				GSXML.PLUGIN_ELEM+GSXML.LIST_MODIFIER, GSXML.PLUGIN_ELEM);
		if(v != null) { // v contains individual <service></service> elements
			pluginList = new PluginData[v.size()];
			for(int i = 0; i < pluginList.length; i++)
				pluginList[i] = new PluginData((Element)v.get(i));
			v.clear();
		}
		v = null;
	}		
	
	/** Useful when adding CollectionData objects to a JList or JCombobox.
	 * @return the name of the collection represented by this CollectionData */
	public String toString() { 
		// return (displayName==null) ? name : displayName; 
		return name;
	}
	
	/** @return a String displaying the contents of this CollectionData object. 
	 * Useful for debugging purposes.*/
	public String show() {
		StringBuffer buf = new StringBuffer("name: " + name);
		buf.append(" -displayname: " + displayName);
		buf.append(" -smallIcon: " + smallIcon);
		buf.append(" -icon: " + icon);
		buf.append(" -url: " + url);
		buf.append("\nColl description: " + description + "\n");
		
		if(serviceList != null) {
			buf.append("\n ServiceList:");
			for(int i = 0; i < serviceList.length; i++) {
				buf.append("\n service:");
				buf.append(" " + serviceList[i].show());
			}
		}
		if(metadataList != null) {
			buf.append("\n MetadataList:");
			for(int i = 0; i < metadataList.length; i++) {
				buf.append("\n metadata:");
				buf.append(" " + metadataList[i].show());
			}
		}
		if(pluginList != null) {
			buf.append("\n PluginList:");
			for(int i = 0; i < pluginList.length; i++) {
				buf.append("\n plugin:");
				buf.append(" " + pluginList[i].show());
			}		
		}
		return buf.toString();
	}
	
	/** @return the Info string for the collection reprsented by this 
	 * CollectionData object. */
	public String info() { return displayName + "\n" + description; }
	
	/* accessor functions - to access the member fields of this CollectionData */
	public String getName() { return name; }
	public String getDisplayName() { return displayName; }
	public String getDescription() { return description; }
	public String getSmallIcon() { return smallIcon; }
	public String getIcon() { return icon; }
	public String getUrl(){ return url; }
	
	public MetaData[] getMetadataList() { return metadataList; }
	public PluginData[] getPluginList() { return pluginList; }

	/** @return the list of services supported by this collection.
	 * @param executableServicesOnly - if true, only the list of supported query  
	 * and browse services are returned. Otherwise, all the services in the 
	 * collection are returned. */
	public ServiceData[] getServiceList(boolean executableServicesOnly) {
		if(executableServicesOnly) { // return only the query and browse services
			ServiceData[] executableServices = null;
			Vector v = new Vector(serviceList.length);
			for(int i = 0; i < serviceList.length; i++) {
				if(serviceList[i].type.equals(GSXML.SERVICE_TYPE_QUERY)
						|| serviceList[i].type.equals(GSXML.SERVICE_TYPE_BROWSE))
				v.add(serviceList[i]);
			}
			executableServices = new ServiceData[v.size()];
			v.toArray(executableServices);
			return executableServices;
		} else // all services
			return serviceList;
	}
	
	/**
	 * Only the name and type of a &lt;service&gt;&lt;/service&gt; element is 
	 * stored, since this is all the service data available inside a collection's
	 * response message to a describe request. And we are here dealing with 
	 * just a CollectionData's service information (hence it's an inner class: 
	 * CollectionData.ServiceData).
	 */
	public static class ServiceData {
		/** name of the Greenstone service represented by this ServiceData object */
		public final String name;
		/** type attribute of &lt;service&gt;&lt;/service&gt; element.
		 * type is optional and therefore can be empty str. */
		public final String type; 
		
		/**Gets a &lt;service&gt;&lt;/service&gt; element and sets its members using 
		 * info extracted from it
		 * @param serviceTag is the &lt;service&gt;&lt;/service&gt; element from
		 * which this ServiceData object is constructed. */
		public ServiceData(Element serviceTag){
			// we'll surely have a name attribute, but to be on the safe side:
			this.name = serviceTag.hasAttribute(GSXML.NAME_ATT) ? 
					serviceTag.getAttribute(GSXML.NAME_ATT) : "";
			// Look for any lang attributes
			this.type = serviceTag.hasAttribute(GSXML.TYPE_ATT) ?
				serviceTag.getAttribute(GSXML.TYPE_ATT) : "";
		}
		
		/** Useful when adding ServiceData objects to a JList or JCombobox
		 * @return the name of this ServiceData, this corresponds to the name of
		 * the Greenstone Service that this object represents. */
		public String toString() { return name; }
		
		/** @return a String displaying the contents of this ServiceData object. 
		 * Useful for debugging purposes.*/
		public String show() {
			StringBuffer buf = new StringBuffer("name: " + name);
			if(!type.equals("")) 
				buf.append(" -type: " + type);
			buf.append("\n");
			return buf.toString();
		}
	}
	
	/** Static inner class MetaData: can import it into other files as
	 * "import gs3client.CollectionData.MetaData;"
	 * In this way, can use it as if it were a regular class (a.o.t. an inner 
	 * class). That is, don't need to prepend the outer-classname if imported
	 * like this.
	 * gs3client.MetaData, from which this class inherits, has name, value and 
	 * bodyText fields. This class adds an additional language field that is 
	 * sometimes present in a Collection's metadata information. So we include 
	 * the language as a member variable here in case we ever need it. 
	*/
	public static class MetaData extends org.greenstone.gs3client.data.MetaData{
		public final String language;
		
		/** Given a &lt;metadata&gt;&lt;/metadata&gt; element, constructs a 
		 * CollectionData sets its members using info extracted from it 
		 * @param metadataTag is the the &lt;metadata&gt;&lt;/metadata&gt; element 
		 * whose data is used to construct this MetaData object. */
		public MetaData(Element metadataTag){
			super(metadataTag);
			// Look for any lang attributes
			this.language = metadataTag.hasAttribute("lang") ?
				metadataTag.getAttribute("lang") : "";
		}
		
		/** @return a String displaying the contents of this MetaData object. 
		 * Useful for debugging purposes.*/
		public String show() {
			StringBuffer buf = new StringBuffer(super.show());
			if(!language.equals("")) 
				buf.append(" -lang: " + language);
			buf.append("\n");
			return buf.toString();
		}
	}

	/** PluginData inner class of CollectionData, represents the data stored in 
	 * a &lt;plugin&gt;&lt;/plugin&gt; element. This class just stores the 
	 * Plugin's name, as well as option subelements (in variable optionnames). */
	public static class PluginData {
		/** name of the plugin represented by this PluginData object */
		public final String name;
		/** A plugin's option subelements. Variable optionnames can be empty 
		 * (it's an optional variable). */
		public final String[] optionnames;
		
		/** Gets a &lt;plugin&gt;&lt;/plugin&gt; element and sets its members 
		 * using info extracted from it. 
		 * @param pluginTag is the the &lt;plugin&gt;&lt;/plugin&gt; element 
		 * whose data is used to construct this PluginData object. */
		public PluginData(Element pluginTag){
			this.name = pluginTag.hasAttribute(GSXML.NAME_ATT) ?
					pluginTag.getAttribute(GSXML.NAME_ATT) : "";
			// Look for any option subElements
			NodeList nl = pluginTag.getElementsByTagName("option");
			if(nl == null || nl.getLength() == 0) optionnames = null;
			else {
				optionnames = new String[nl.getLength()];
				for(int i = 0; i < optionnames.length; i++) {
					Element option = (Element)nl.item(i);
					optionnames[i] = option.getAttribute(GSXML.NAME_ATT);
				}
			}
		}

		/** Useful when adding PluginData objects to a JList or JCombobox. 
		 * @return the name of the plugin represented by this PluginData object. */
		public String toString() { return name; }
		
		/** @return a String displaying the contents of this PluginData object. 
		 * Useful for debugging purposes. */
		public String show() {
			StringBuffer buf = new StringBuffer("name: " + name);
			if(optionnames != null){
				buf.append("\n  optionnames: ");
				for(int i = 0; i < optionnames.length; i++)
					buf.append(optionnames[i] + " ");
			}
			buf.append("\n");
			return buf.toString();
		}
	}
}